open ExtString;;
open ExtList;;
exception Quit;;
external set_tcp_nodelay : Unix.file_descr -> bool -> unit = "stub_set_tcp_nodelay";;
let curtime () = Unix.gettimeofday ();;
let pi = 3.14159265;;
let (|>) x f = f x;;

let rec select sock delay =
  let t0 = curtime () in
  try let (rlst,_,_) = Unix.select [sock] [] [] delay in rlst
  with
  | Unix.Unix_error(Unix.EAGAIN,_,_)
	| Unix.Unix_error(Unix.EINTR, _,_) ->
    let newdelay = delay -. (curtime()-.t0) in
    select sock (if newdelay > 0.0 then newdelay else 0.0)
  | ex -> raise ex;;

let connect host port =
  let addr = 
    try let h = Unix.gethostbyname host in 
      if Array.length h.Unix.h_addr_list > 0 then 
			  h.Unix.h_addr_list.(0) else raise Not_found
    with Not_found-> try Unix.inet_addr_of_string host
    with Failure _ -> 
     Printf.printf "Cannot resolve %s with neither gethostbyname nor inet_addr_of_string\n" host;
     raise Quit
  in let sa = Unix.ADDR_INET(addr, port) in
	let sock = Unix.socket Unix.PF_INET Unix.SOCK_STREAM 0 in
	Unix.connect sock sa;
  Unix.set_nonblock sock;
	set_tcp_nodelay sock true;
	sock;;

let msg_buf = Buffer.create 10240;;

let receive onmsg fd = 
  let buflen = 10240 in let buf = String.create buflen in
 	let rec readloop () = 
		let n = Unix.recv fd buf 0 buflen [] in 
    Buffer.add_substring msg_buf buf 0 n; 
    if n > 0 then readloop()
  in try readloop () with 
  	| Unix.Unix_error(Unix.EAGAIN,_,_)
		| Unix.Unix_error(Unix.EWOULDBLOCK,_,_)->
  try 
		let rec call_onmsg = function
      | [] -> failwith "msglst is empty"
      | last :: [] -> Buffer.reset msg_buf; Buffer.add_string msg_buf last
      | msg :: tl -> onmsg msg; call_onmsg tl
    in call_onmsg (String.nsplit (Buffer.contents msg_buf) ";")
  with Invalid_string->();;

type world_data = { mutable dx : float; mutable dy : float;
  mutable min_sensor : float; mutable max_sensor : float;
  mutable max_speed : float; 
  mutable max_turn : float; mutable max_hard_turn : float
};;

let limits = { dx = 0.0; dy = 0.0; min_sensor = 0.0;
 max_sensor = 0.0; max_speed = 0.0; max_turn = 0.0; max_hard_turn = 0.0};; 

type rover_state = { time : float; 
  ctl : string; mx : float; my : float;
  dir : float; speed : float; rotspd : float };;

let rover = ref { time = -1.0;  ctl = "--"; mx = 1.0; my = 1.0; 
                  dir = 0.0; speed = 0.0; rotspd = 0.0};;
let rover0 = ref { time = -1.0;  ctl = "--"; mx = 1.0; my = 1.0; 
                  dir = 0.0; speed = 0.0; rotspd = 0.0};;

let skip_learn() = 
  rover0 := {!rover0 with time = -1.0};
  rover := {!rover with time = -1.0};;

type obj_type = Boulder | Crater | Home;;
type map_object = obj_type * float;;

let trn = Hashtbl.create 4000;;

let angdiff a1 a0 =
  let d = a1 -. a0 in 
  if d <= -180.0 then d +. 360.0 else 
  if d >= 180.0 then d -. 360.0 else d;;

let mrt = ref [];;

let parse_msg s = 
  if s="" then () else
  let rec parse_lst = function  
    | "I" :: dx :: dy :: timlim :: minsens :: maxsens :: maxspd 
        :: maxturn :: maxhardturn :: tl -> 
      limits.dx <- (float_of_string dx) /. 2.0;
      limits.dy <- (float_of_string dy) /. 2.0;
      limits.min_sensor <- float_of_string minsens;
      limits.max_sensor <- float_of_string maxsens;
      limits.max_speed <- float_of_string maxspd;
      limits.max_turn <- (float_of_string maxturn) /. 180.0 *. pi;
      limits.max_hard_turn <- (float_of_string maxhardturn) /. 180.0 *. pi
    | "T" :: tm :: ctl :: x :: y :: dir :: spd :: tl ->
      let fdir = float_of_string dir in
      let ftime = float_of_string tm in
      let rov = { time = ftime; ctl = ctl;
                  mx = float_of_string x; my = float_of_string y;
                  dir = fdir; speed = float_of_string spd;
                  rotspd = if !rover.time < 0.0 then 0.0 else 
                           let dt = (ftime -. !rover.time) /. 1000.0 in
                           (angdiff fdir !rover.dir) /. dt } in
      rover0 := !rover;
      rover := rov;
      mrt := [];
      parse_lst tl
    | "b" :: x :: y :: r :: tl ->
      let coord = (float_of_string x, float_of_string y) in
      Hashtbl.replace trn coord (Boulder, 0.5 +. float_of_string r);
      parse_lst tl
    | "c" :: x :: y :: r :: tl ->
      let coord = (float_of_string x, float_of_string y) in
      Hashtbl.replace trn coord (Crater, float_of_string r);
      parse_lst tl
    | "h" :: x :: y :: r :: tl ->
      parse_lst tl
    | "m" :: x :: y :: dir :: spd :: tl ->
      mrt := { time = 0.0; ctl = "a-"; 
               mx = float_of_string x; my = float_of_string y;
               dir = float_of_string dir; speed = float_of_string spd;
               rotspd = 0.0 }  :: !mrt;
      parse_lst tl
    | "B" :: tl 
    | "E" :: tl -> skip_learn() 
    | _ -> ()
  in parse_lst (String.nsplit s " ");;

type near_obj = { ox : float; oy : float; obr : float; minr : float };;
type nearest_two = { all : near_obj; down : near_obj };;

let mkloctrn x y =
  let range = (limits.max_sensor +. limits.min_sensor) *. 2.0 in
  let loctrn = Hashtbl.create 50 in
  Hashtbl.iter (fun ((ox,oy) as ocrd) knd_obr-> 
    let d = abs_float (ox -. x) +. abs_float (oy -. y) in
    if d < range then Hashtbl.replace loctrn ocrd knd_obr) trn;
  loctrn;;

let mkfield loctrn mlst =
  let maxheight = limits.dx +. limits.dy in 
  let field x y curspd dir =
    let rxy = sqrt (x*.x +. y*.y) in
    let rdir = dir /. 180.0 *. pi in
    let vx = curspd *. cos rdir 
    and vy = curspd *. sin rdir in
    let calch (ox,oy) (knd,obr) maxh =
      let dx = ox -. x and dy = oy -. y in
      let dist = sqrt (dx*.dx +. dy*.dy) in
      let r = dist -. obr in 
      if dist = 0.0 then maxheight *. 2.0 else
      let vo = (vx *. dx +. vy *. dy) /. dist in
      let h  = 
        if vo > 0.0 then
          let turn_radius = vo /. limits.max_hard_turn in
          let turn_dist = sqrt (obr*.(obr +. turn_radius *. 2.0)) in
          if r <= turn_dist then rxy +. turn_dist -. r else rxy
        else rxy
      in max h maxh
    in let objh = Hashtbl.fold calch loctrn 0.0 in
    List.fold_right 
      (fun m maxh-> calch (m.mx, m.my) (Boulder, 1.2) maxh) mlst objh
  in field;;

let cmd_manual = function
  | '7' -> "al;" | '8' -> "a;" | '9' -> "ar;"
  | '4' -> "l;" | '5' -> ";" | '6' -> "r;"
  | '1' -> "bl;" | '2' -> "b;" | '3' -> "br;"
  | 'q' -> raise Quit | _ -> "";;

type physics_info = { mutable acc : float; mutable brk : float; 
  mutable k : float;  mutable turnacc : float };;
let ph = { acc = 0.0; brk = 0.0; k = 0.0; turnacc = 0.0; };;

type acc_witness = {mutable a1 : float; mutable a2 : float;
                 mutable v1 : float; mutable v2 : float};; (*v1<v2*)
let accwitn = { a1 = 0.0; a2 = 0.0; v1 = 0.0; v2 = 0.0};;
let brkwitn = { a1 = 0.0; a2 = 0.0; v1 = 0.0; v2 = 0.0};;

let learn_acc a v aw =
  if v > aw.v2 then (aw.v2 <- v; aw.a2 <- a);
  if v < aw.v1 || aw.v1=0.0 then (aw.v1 <- v; aw.a1 <- a);
  if aw.v1 <> aw.v2 then
    let k = (aw.a1 -. aw.a2) /. (aw.v2*.aw.v2 -. aw.v1*.aw.v1) in
    let acc = aw.a1 +. k *. aw.v1 *. aw.v1 in Some (acc,k)
  else None;;

let learn () =
  (if !rover0.time < 0.0 then () else
  let spdstate = !rover.ctl.[0] in
  let dt = (!rover.time -. !rover0.time) /. 1000.0 in
  if dt = 0.0 then () else
  ((match spdstate with
  | 'a'-> (let a = (!rover.speed -. !rover0.speed) /. dt  in
          match learn_acc a !rover0.speed accwitn with
          | None->()
          | Some(acc,k)-> ph.acc <- ph.acc *. 0.75 +. acc *. 0.25;
                          ph.k <- ph.k *. 0.75 +. k *. 0.25)
  | 'b'-> (if !rover.speed = 0.0 then () else
          let a = (!rover.speed -. !rover0.speed) /. dt in
          match learn_acc a !rover0.speed brkwitn with
          | None->()
          | Some(brk,k)-> ph.brk <- ph.brk *. 0.75 +. brk *. 0.25;
                          ph.k <- ph.k *. 0.75 +. k *. 0.25)
  |  _ -> ());
  let arot = abs_float (!rover.rotspd -. !rover0.rotspd) /. dt in
  if arot >= 0.1 then ph.turnacc <- max ph.turnacc arot));;

let nextspeedstate st c = (* 'a' 'b' => '-' *)
  match st,c with
  |  _ , '-' -> st
  | 'a', 'a' -> 'a'
  | 'a', 'b' -> '-'
  | '-', 'a' -> 'a'
  | '-', 'b' -> 'b'
  | 'b', 'a' -> '-'
  | 'b', 'b' -> 'b'
  | _, _ -> failwith "nextspeedstate bad param";;

let nextturnstate st c =
  match st, c with
  | _, '-' -> st
  | 'L', 'l' -> 'L'
  | 'L', 'r' -> 'l'
  | 'l', 'l' -> 'L'
  | 'l', 'r' -> '-'
  | '-', 'l' -> 'l'
  | '-', 'r' -> 'r'
  | 'r', 'l' -> '-'
  | 'r', 'r' -> 'R'
  | 'R', 'l' -> 'r'
  | 'R', 'r' -> 'R'
  | _, _ -> failwith "nextturnstate bad param";;

let nextstate st cmd = (* "a-" "r;" => "ar" *)
  let acmd = match cmd.[0] with 'a'->'a'|'b'->'b'|_->'-' in
  let tcmd = if String.contains cmd 'l' then 'l' else
             if String.contains cmd 'r' then 'r' else '-' in
  String.implode [nextspeedstate st.[0] acmd; nextturnstate st.[1] tcmd];;

let nextplace ctl dt rov =
  let a = match ctl.[0] with
    | 'a' -> if ph.acc<>0.0 then ph.acc else 2.0
    | 'b' -> if ph.brk<>0.0 then ph.brk else -2.0
    |  _  -> 0.0 in
  let maxrotspd = match ctl.[1] with
    | 'l' | 'r' -> limits.max_turn *. 180.0 /. pi
    | 'L' | 'R' -> limits.max_hard_turn *. 180.0 /. pi
    |  _ -> 1.0 in
  let ta = if ph.turnacc<>0.0 then ph.turnacc else 80.0 in
  let mkarot spd =
    if rov.rotspd > spd +. 1.0 then (-.ta) else
    if rov.rotspd < spd -. 1.0 then ta else 0.0 in
  let arot = match ctl.[1] with
    | 'L' -> mkarot maxrotspd 
    | 'l' -> mkarot maxrotspd 
    | 'R' -> mkarot (-.maxrotspd) 
    | 'r' -> mkarot (-.maxrotspd) 
    |  _  -> mkarot 0.0 in
  let nspeed = rov.speed +. max (a*.dt -. ph.k*.rov.speed*.rov.speed*.dt) 0.0 in
  let s = (rov.speed +. nspeed) *. 0.5 *. dt in  
  let nrotspd = rov.rotspd +. arot *.dt in
  let w = (rov.rotspd +. nrotspd) *. 0.5 in
  let ndir = let dr = rov.dir +. w *. dt in
             if dr <= -.180.0 then dr +. 360.0 else
             if dr > 180.0 then dr -. 360.0 else dr in
  let raddir = ndir /. 180.0 *. pi in
  let dx = s *. cos raddir 
  and dy = s *. sin raddir in
  { time = rov.time +. dt;  ctl = ctl; 
    mx = rov.mx +. dx; my = rov.my +. dy; 
    dir = ndir; speed = nspeed; rotspd = nrotspd};;

let gottabreak rov =
  let rdir = rov.dir /. 180.0 *. pi in
  let vx = rov.speed *. cos rdir 
  and vy = rov.speed *. sin rdir in
  let r = sqrt(rov.mx *. rov.mx +. rov.my *. rov.my) in
  let cx = rov.mx /. r and cy = rov.my /. r in
  let vt = abs_float (vx *. cy -. vy *. cx) in
  vt >= limits.max_hard_turn *. r *. 0.5;;

let movemar mlst dt =
  List.map (fun m->
    let rdir = m.dir /. 180.0 *. pi in
    let vx = m.speed *. cos rdir 
    and vy = m.speed *. sin rdir in
    { m with mx = m.mx +. vx *. dt; my = m.my +. vy *. dt }) mlst;; 

let think () = 
  learn ();
  let loctrn = mkloctrn !rover.mx !rover.my in
  let cmds = 
    if gottabreak !rover then ["b;"; "bl;"; "br;"]
    else ["a;"; "al;"; "ar;"; "l;"; "r;"; ";"; "b;"; "bl;"; "br;"] in
  let maxheight = limits.dx +. limits.dy in
  let marsteps = Array.of_list 
    [movemar !mrt 0.1; movemar !mrt 0.2; movemar !mrt 0.3] in
  let rec findbest rov steps = (* returns (cmd, h) *)
    let field = mkfield loctrn marsteps.(3-steps) in
    let cmpcmd cmd ((bestcmd, besth) as best) =
      let nctl = nextstate rov.ctl cmd and mdt = 0.025 in
      let nrov = rov |> nextplace nctl mdt |> nextplace nctl mdt 
        |> nextplace nctl mdt |> nextplace nctl mdt  in
      let th = field nrov.mx nrov.my nrov.speed nrov.dir in
      if th >= maxheight then best else
      let h = if steps=1 then th else snd (findbest nrov (steps-1)) in
      if h < besth then (cmd, h) else best
    in 
    List.enum cmds |> Enum.fold cmpcmd (";", maxheight*.2.0)
  in fst (findbest !rover 3);;

let main host port = 
  try let sock = connect host port in 
  let receive_fun = receive parse_msg in
	let rec loop () =
	  let rlst = select sock 0.100 in	List.iter receive_fun rlst;
    let cmd = think() in
		ignore(Unix.write sock cmd 0 (String.length cmd));
    loop ()
	in (try loop () with 
  | Unix.Unix_error(Unix.ECONNRESET,_,_)
  | Quit -> ()
  | ex   -> Printf.printf "Exception: %s\n" (Printexc.to_string ex));
  Unix.close sock;
  with Quit -> ()
  | Unix.Unix_error(_, "connect",_) -> 
      Printf.printf "Cannot connect to %s:%d\n" host port;;

if Array.length Sys.argv < 3 then print_endline "run hostname port" else
main Sys.argv.(1) (int_of_string Sys.argv.(2));;